#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional
from urllib.parse import urlencode

import pendulum
import requests
from airbyte_cdk.sources.streams.core import package_name_from_class
from airbyte_cdk.sources.utils import casing
from airbyte_cdk.sources.utils.schema_helpers import ResourceSchemaLoader
from airbyte_protocol.models import SyncMode
from source_linkedin_ads.streams import Campaigns, Creatives, IncrementalLinkedinAdsStream

from .utils import get_parent_stream_values, transform_data

# Number of days ahead for date slices, from start date.
WINDOW_IN_DAYS = 30
# List of Reporting Metrics fields available for fetch
ANALYTICS_FIELDS_V2: List = [
    "actionClicks",
    "adUnitClicks",
    "approximateUniqueImpressions",
    "cardClicks",
    "cardImpressions",
    "clicks",
    "commentLikes",
    "comments",
    "companyPageClicks",
    "conversionValueInLocalCurrency",
    "costInLocalCurrency",
    "costInUsd",
    "dateRange",
    "documentCompletions",
    "documentFirstQuartileCompletions",
    "documentMidpointCompletions",
    "documentThirdQuartileCompletions",
    "downloadClicks",
    "externalWebsiteConversions",
    "externalWebsitePostClickConversions",
    "externalWebsitePostViewConversions",
    "follows",
    "fullScreenPlays",
    "impressions",
    "jobApplications",
    "jobApplyClicks",
    "landingPageClicks",
    "leadGenerationMailContactInfoShares",
    "leadGenerationMailInterestedClicks",
    "likes",
    "oneClickLeadFormOpens",
    "oneClickLeads",
    "opens",
    "otherEngagements",
    "pivotValues",
    "postClickJobApplications",
    "postClickJobApplyClicks",
    "postClickRegistrations",
    "postViewJobApplications",
    "postViewJobApplyClicks",
    "postViewRegistrations",
    "reactions",
    "registrations",
    "sends",
    "shares",
    "talentLeads",
    "textUrlClicks",
    "totalEngagements",
    "validWorkEmailLeads",
    "videoCompletions",
    "videoFirstQuartileCompletions",
    "videoMidpointCompletions",
    "videoStarts",
    "videoThirdQuartileCompletions",
    "videoViews",
    "viralCardClicks",
    "viralCardImpressions",
    "viralClicks",
    "viralCommentLikes",
    "viralComments",
    "viralCompanyPageClicks",
    "viralDocumentCompletions",
    "viralDocumentFirstQuartileCompletions",
    "viralDocumentMidpointCompletions",
    "viralDocumentThirdQuartileCompletions",
    "viralDownloadClicks",
    "viralExternalWebsiteConversions",
    "viralExternalWebsitePostClickConversions",
    "viralExternalWebsitePostViewConversions",
    "viralFollows",
    "viralFullScreenPlays",
    "viralImpressions",
    "viralJobApplications",
    "viralJobApplyClicks",
    "viralLandingPageClicks",
    "viralLikes",
    "viralOneClickLeadFormOpens",
    "viralOneClickLeads",
    "viralOtherEngagements",
    "viralPostClickJobApplications",
    "viralPostClickJobApplyClicks",
    "viralPostClickRegistrations",
    "viralPostViewJobApplications",
    "viralPostViewJobApplyClicks",
    "viralPostViewRegistrations",
    "viralReactions",
    "viralRegistrations",
    "viralShares",
    "viralTotalEngagements",
    "viralVideoCompletions",
    "viralVideoFirstQuartileCompletions",
    "viralVideoMidpointCompletions",
    "viralVideoStarts",
    "viralVideoThirdQuartileCompletions",
    "viralVideoViews",
]


class LinkedInAdsAnalyticsStream(IncrementalLinkedinAdsStream, ABC):
    """
    AdAnalytics Streams more info:
    https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads-reporting/ads-reporting?tabs=curl&view=li-lms-2023-05#analytics-finder
    """

    endpoint = "adAnalytics"
    # For Analytics streams, the primary_key is the entity of the pivot [Campaign URN, Creative URN, etc.] + `end_date`
    primary_key = ["string_of_pivot_values", "end_date"]
    cursor_field = "end_date"
    records_limit = 15000
    FIELDS_CHUNK_SIZE = 18

    def get_json_schema(self) -> Mapping[str, Any]:
        schema = ResourceSchemaLoader(package_name_from_class(self.__class__)).get_schema("ad_analytics")
        schema["properties"].update({self.search_param_value: {"type": ["null", "string"]}})
        return schema

    def __init__(self, name: str = None, pivot_by: str = None, time_granularity: str = None, **kwargs):
        self.user_stream_name = name
        if pivot_by:
            self.pivot_by = pivot_by
        if time_granularity:
            self.time_granularity = time_granularity
        super().__init__(**kwargs)

    @property
    @abstractmethod
    def search_param(self) -> str:
        """
        :return: Search parameters for the request
        """

    @property
    @abstractmethod
    def search_param_value(self) -> str:
        """
        :return: Name field to filter by
        """

    @property
    @abstractmethod
    def parent_values_map(self) -> Mapping[str, str]:
        """
        :return: Mapping for parent child relation
        """

    @property
    def name(self) -> str:
        """We override the stream name to let the user change it via configuration."""
        name = self.user_stream_name or self.__class__.__name__
        return casing.camel_to_snake(name)

    @property
    def base_analytics_params(self) -> MutableMapping[str, Any]:
        """Define the base parameters for analytics streams"""
        return {"q": "analytics", "pivot": f"(value:{self.pivot_by})", "timeGranularity": f"(value:{self.time_granularity})"}

    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        headers = super().request_headers(stream_state, stream_slice, next_page_token)
        return headers | {"X-Restli-Protocol-Version": "2.0.0"}

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        params = self.base_analytics_params
        params.update(**self.update_analytics_params(stream_slice))
        params[self.search_param] = f"List(urn%3Ali%3A{self.search_param_value}%3A{self.get_primary_key_from_slice(stream_slice)})"
        return urlencode(params, safe="():,%")

    @staticmethod
    def update_analytics_params(stream_slice: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        Produces the date range parameters from input stream_slice
        """
        date_range = stream_slice["dateRange"]
        return {
            "dateRange": f"(start:(year:{date_range['start.year']},month:{date_range['start.month']},day:{date_range['start.day']}),"
            f"end:(year:{date_range['end.year']},month:{date_range['end.month']},day:{date_range['end.day']}))",
            # Chunk of fields
            "fields": stream_slice["fields"],
        }

    def next_page_token(self, response: requests.Response) -> Optional[Mapping[str, Any]]:
        """
        Pagination is not supported
        (See Restrictions: https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads-reporting/ads-reporting?view=li-lms-2023-09&tabs=http#restrictions)
        """
        parsed_response = response.json()
        is_elements_less_than_limit = len(parsed_response.get("elements")) < self.records_limit

        # Note: The API might return fewer records than requested within the limits during pagination.
        # This behavior is documented at: https://github.com/airbytehq/airbyte/issues/34164
        paging_params = parsed_response.get("paging", {})
        is_end_of_records = (
            paging_params["total"] - paging_params["start"] <= self.records_limit
            if all(param in paging_params for param in ("total", "start"))
            else True
        )

        if is_elements_less_than_limit and is_end_of_records:
            return None
        raise Exception(
            f"Limit {self.records_limit} elements exceeded. "
            f"Try to request your data in more granular pieces. "
            f"(For example switch `Time Granularity` from MONTHLY to DAILY)"
        )

    def get_primary_key_from_slice(self, stream_slice) -> str:
        return stream_slice.get(self.primary_slice_key)

    def stream_slices(
        self, *, sync_mode: SyncMode, cursor_field: Optional[List[str]] = None, stream_state: Optional[Mapping[str, Any]] = None
    ) -> Iterable[Optional[Mapping[str, List[Mapping[str, Any]]]]]:
        """
        LinkedIn has a max of 20 fields per request. We make chunks by size of 19 fields to have the `dateRange` be included as well.
        https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads-reporting/ads-reporting?view=li-lms-2023-05&tabs=http#requesting-specific-metrics-in-the-analytics-finder

        :param sync_mode:
        :param cursor_field:
        :param stream_state:
        :return: An iterable of dictionaries, each containing a single key 'field_date_chunks'. The value under 'field_date_chunks' is
        a list of dictionaries where each dictionary represents a slice of data defined by a specific date range and chunked fields.

        Example of returned data:
        {
            'field_date_chunks': [
                {
                    'campaign_id': 123,
                    'fields': 'field_1,field_2,dateRange',
                    'dateRange': {
                        'start.day': 1, 'start.month': 1, 'start.year': 2020,
                        'end.day': 30, 'end.month': 1, 'end.year': 2020
                    }
                },
                {
                    'campaign_id': 123,
                    'fields': 'field_3,field_4,dateRange',
                    'dateRange': {
                        'start.day': 1, 'start.month': 1, 'start.year': 2020,
                        'end.day': 30, 'end.month': 1, 'end.year': 2020
                    }
                }
            ]
        }
        """
        parent_stream = self.parent_stream(config=self.config)
        stream_state = stream_state or {self.cursor_field: self.config.get("start_date")}
        for record in parent_stream.read_records(sync_mode=sync_mode):
            base_slice = get_parent_stream_values(record, self.parent_values_map)
            for date_slice in self.get_date_slices(stream_state.get(self.cursor_field), self.config.get("end_date")):
                date_slice_with_fields: List = []
                for fields_set in self.chunk_analytics_fields():
                    base_slice["fields"] = ",".join(fields_set)
                    date_slice_with_fields.append(base_slice | date_slice)
                yield {"field_date_chunks": date_slice_with_fields}

    @staticmethod
    def get_date_slices(start_date: str, end_date: str = None, window_in_days: int = WINDOW_IN_DAYS) -> Iterable[Mapping[str, Any]]:
        """
        Produces date slices from start_date to end_date (if specified),
        otherwise end_date will be present time.
        """
        start = pendulum.parse(start_date)
        end = pendulum.parse(end_date) if end_date else pendulum.now()
        date_slices = []
        while start < end:
            slice_end_date = start.add(days=window_in_days)
            date_slice = {
                "start.day": start.day,
                "start.month": start.month,
                "start.year": start.year,
                "end.day": slice_end_date.day,
                "end.month": slice_end_date.month,
                "end.year": slice_end_date.year,
            }
            date_slices.append({"dateRange": date_slice})
            start = slice_end_date
        yield from date_slices

    @staticmethod
    def chunk_analytics_fields(
        fields: List = ANALYTICS_FIELDS_V2,
        fields_chunk_size: int = FIELDS_CHUNK_SIZE,
    ) -> Iterable[List]:
        """
        Chunks the list of available fields into the chunks of equal size.
        """
        # Make chunks
        chunks = list((fields[f : f + fields_chunk_size] for f in range(0, len(fields), fields_chunk_size)))
        # Make sure base_fields are within the chunks
        for chunk in chunks:
            if "dateRange" not in chunk:
                chunk.append("dateRange")
            if "pivotValues" not in chunk:
                chunk.append("pivotValues")
        yield from chunks

    def read_records(
        self, stream_state: Mapping[str, Any] = None, stream_slice: Optional[Mapping[str, Any]] = None, **kwargs
    ) -> Iterable[Mapping[str, Any]]:
        merged_records = defaultdict(dict)
        for field_slice in stream_slice.get("field_date_chunks", []):
            for rec in super().read_records(stream_slice=field_slice, **kwargs):
                merged_records[f"{rec[self.cursor_field]}-{rec['pivotValues']}"].update(rec)
        yield from merged_records.values()

    def parse_response(self, response: requests.Response, **kwargs) -> Iterable[Mapping]:
        """
        We need to get out the nested complex data structures for further normalization, so the transform_data method is applied.
        """
        for rec in transform_data(response.json().get("elements")):
            yield rec | {self.search_param_value: self.get_primary_key_from_slice(kwargs.get("stream_slice")), "pivot": self.pivot_by}


class AdCampaignAnalytics(LinkedInAdsAnalyticsStream):
    """
    Campaign Analytics stream.
    """

    endpoint = "adAnalytics"

    parent_stream = Campaigns
    parent_values_map = {"campaign_id": "id"}
    search_param = "campaigns"
    search_param_value = "sponsoredCampaign"
    pivot_by = "CAMPAIGN"
    time_granularity = "DAILY"


class AdCreativeAnalytics(LinkedInAdsAnalyticsStream):
    """
    Creative Analytics stream.
    """

    parent_stream = Creatives
    parent_values_map = {"creative_id": "id"}
    search_param = "creatives"
    search_param_value = "sponsoredCreative"
    pivot_by = "CREATIVE"
    time_granularity = "DAILY"

    def get_primary_key_from_slice(self, stream_slice) -> str:
        creative_id = stream_slice.get(self.primary_slice_key).split(":")[-1]
        return creative_id


class AdImpressionDeviceAnalytics(AdCampaignAnalytics):
    pivot_by = "IMPRESSION_DEVICE_TYPE"


class AdMemberCompanySizeAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_COMPANY_SIZE"


class AdMemberIndustryAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_INDUSTRY"


class AdMemberSeniorityAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_SENIORITY"


class AdMemberJobTitleAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_JOB_TITLE"


class AdMemberJobFunctionAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_JOB_FUNCTION"


class AdMemberCountryAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_COUNTRY_V2"


class AdMemberRegionAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_REGION_V2"


class AdMemberCompanyAnalytics(AdCampaignAnalytics):
    pivot_by = "MEMBER_COMPANY"



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import logging
from typing import Any, List, Mapping, Optional, Tuple, Union

from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.http.requests_native_auth import Oauth2Authenticator, TokenAuthenticator
from airbyte_cdk.utils import AirbyteTracedException
from airbyte_protocol.models import FailureType
from source_linkedin_ads.analytics_streams import (
    AdCampaignAnalytics,
    AdCreativeAnalytics,
    AdImpressionDeviceAnalytics,
    AdMemberCompanyAnalytics,
    AdMemberCompanySizeAnalytics,
    AdMemberCountryAnalytics,
    AdMemberIndustryAnalytics,
    AdMemberJobFunctionAnalytics,
    AdMemberJobTitleAnalytics,
    AdMemberRegionAnalytics,
    AdMemberSeniorityAnalytics,
)
from source_linkedin_ads.streams import Accounts, AccountUsers, CampaignGroups, Campaigns, Conversions, Creatives

logger = logging.getLogger("airbyte")


class SourceLinkedinAds(AbstractSource):
    """
    Abstract Source inheritance, provides:
    - implementation for `check` connector's connectivity
    - implementation to call each stream with it's input parameters.
    """

    @classmethod
    def get_authenticator(cls, config: Mapping[str, Any]) -> Union[TokenAuthenticator, Oauth2Authenticator]:
        """
        Validate input parameters and generate a necessary Authentication object
        This connectors support 2 auth methods:
        1) direct access token with TTL = 2 months
        2) refresh token (TTL = 1 year) which can be converted to access tokens,
           Every new refresh revokes all previous access tokens
        """
        auth_method = config.get("credentials", {}).get("auth_method")
        if not auth_method or auth_method == "access_token":
            # support of backward compatibility with old exists configs
            access_token = config["credentials"]["access_token"] if auth_method else config["access_token"]
            return TokenAuthenticator(token=access_token)
        elif auth_method == "oAuth2.0":
            return Oauth2Authenticator(
                token_refresh_endpoint="https://www.linkedin.com/oauth/v2/accessToken",
                client_id=config["credentials"]["client_id"],
                client_secret=config["credentials"]["client_secret"],
                refresh_token=config["credentials"]["refresh_token"],
            )
        raise Exception("incorrect input parameters")

    def check_connection(self, logger: logging.Logger, config: Mapping[str, Any]) -> Tuple[bool, Optional[Any]]:
        """
        Testing connection availability for the connector.
        :: for this check method the Customer must have the "r_liteprofile" scope enabled.
        :: more info: https://docs.microsoft.com/linkedin/consumer/integrations/self-serve/sign-in-with-linkedin
        """
        self._validate_ad_analytics_reports(config)
        config["authenticator"] = self.get_authenticator(config)
        stream = Accounts(config)
        try:
            return stream.check_availability(logger)
        except Exception as e:
            return False, e

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        """
        Mapping a input config of the user input configuration as defined in the connector spec.
        Passing config to the streams.
        """
        self._validate_ad_analytics_reports(config)
        config["authenticator"] = self.get_authenticator(config)
        streams = [
            Accounts(config),
            AccountUsers(config),
            AdCampaignAnalytics(config=config),
            AdCreativeAnalytics(config=config),
            AdImpressionDeviceAnalytics(config=config),
            AdMemberCompanySizeAnalytics(config=config),
            AdMemberCountryAnalytics(config=config),
            AdMemberJobFunctionAnalytics(config=config),
            AdMemberJobTitleAnalytics(config=config),
            AdMemberIndustryAnalytics(config=config),
            AdMemberSeniorityAnalytics(config=config),
            AdMemberRegionAnalytics(config=config),
            AdMemberCompanyAnalytics(config=config),
            CampaignGroups(config=config),
            Campaigns(config=config),
            Creatives(config=config),
            Conversions(config=config),
        ]

        return streams + self.get_custom_ad_analytics_reports(config)

    def get_custom_ad_analytics_reports(self, config: Mapping[str, Any]) -> List[Stream]:
        streams = []

        for ad_report in config.get("ad_analytics_reports", []):
            stream = AdCampaignAnalytics(
                name=f"Custom{ad_report.get('name')}",
                pivot_by=ad_report.get("pivot_by"),
                time_granularity=ad_report.get("time_granularity"),
                config=config,
            )
            streams.append(stream)

        return streams

    def _validate_ad_analytics_reports(self, config: Mapping[str, Any]) -> None:
        report_names = [x["name"] for x in config.get("ad_analytics_reports", [])]
        if len(report_names) != len(set(report_names)):
            report_names = [x["name"] for x in config.get("ad_analytics_reports")]
            message = f"Stream names for Custom Ad Analytics reports should be unique, duplicated streams: {set(name for name in report_names if report_names.count(name) > 1)}"
            raise AirbyteTracedException(message=message, failure_type=FailureType.config_error)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Mapping, MutableMapping, Optional
from urllib.parse import quote, urlencode

import pendulum
import requests
from airbyte_cdk.sources.streams.http import HttpStream
from airbyte_cdk.sources.utils.transform import TransformConfig, TypeTransformer

from .utils import get_parent_stream_values, transform_data

logger = logging.getLogger("airbyte")

LINKEDIN_VERSION_API = "202404"


class LinkedinAdsStream(HttpStream, ABC):
    """
    Basic class provides base functionality for all streams.
    """

    url_base = "https://api.linkedin.com/rest/"
    primary_key = "id"
    records_limit = 500
    transformer = TypeTransformer(TransformConfig.DefaultSchemaNormalization)

    def __init__(self, config: Dict):
        super().__init__(authenticator=config.get("authenticator"))
        self.config = config
        self.date_time_fields = self._get_date_time_items_from_schema()

    def _get_date_time_items_from_schema(self):
        """
        Get all properties from schema with format: 'date-time'
        """
        schema = self.get_json_schema()
        return [k for k, v in schema["properties"].items() if v.get("format") == "date-time"]

    @property
    def accounts(self):
        """Property to return the list of the user Account Ids from input"""
        return ",".join(map(str, self.config.get("account_ids", [])))

    @property
    @abstractmethod
    def endpoint(self) -> str:
        """Endpoint associated with the current stream"""

    def path(
        self,
        *,
        stream_state: Mapping[str, Any] = None,
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> str:
        """Returns the API endpoint path for stream, from `endpoint` class attribute."""
        return self.endpoint

    def next_page_token(self, response: requests.Response) -> Optional[Mapping[str, Any]]:
        """
        Cursor based pagination using the pageSize and pageToken parameters.
        """
        parsed_response = response.json()
        if parsed_response.get("metadata", {}).get("nextPageToken"):
            return {"pageToken": parsed_response["metadata"]["nextPageToken"]}
        else:
            return None

    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        return {"Linkedin-Version": LINKEDIN_VERSION_API}

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        params = {"pageSize": self.records_limit, "q": "search"}
        if next_page_token:
            params.update(**next_page_token)
        return params

    def parse_response(self, response: requests.Response, **kwargs) -> Iterable[Mapping]:
        """
        We need to get out the nested complex data structures for further normalization, so the transform_data method is applied.
        """
        for record in transform_data(response.json().get("elements")):
            yield self._date_time_to_rfc3339(record)

    def _date_time_to_rfc3339(self, record: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        """
        Transform 'date-time' items to RFC3339 format
        """
        for item in record:
            if item in self.date_time_fields and record[item]:
                record[item] = pendulum.parse(record[item]).to_rfc3339_string()
        return record

    def should_retry(self, response: requests.Response) -> bool:
        if response.status_code == 429:
            error_message = (
                f"Stream {self.name}: LinkedIn API requests are rate limited. "
                f"Rate limits specify the maximum number of API calls that can be made in a 24 hour period. "
                f"These limits reset at midnight UTC every day. "
                f"You can find more information here https://docs.airbyte.com/integrations/sources/linkedin-ads. "
                f"Also quotas and usage are here: https://www.linkedin.com/developers/apps."
            )
            self.logger.error(error_message)
        return super().should_retry(response)


class OffsetPaginationMixin:
    """Mixin for offset based pagination for endpoints tha tdoesnt support cursor based pagination"""

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        params = {"count": self.records_limit, "q": "search"}
        if next_page_token:
            params.update(**next_page_token)
        return params

    def next_page_token(self, response: requests.Response) -> Optional[Mapping[str, Any]]:
        """
        To paginate through results, begin with a start value of 0 and a count value of N.
        To get the next page, set start value to N, while the count value stays the same.
        We have reached the end of the dataset when the response contains fewer elements than the `count` parameter request.
        https://docs.microsoft.com/en-us/linkedin/shared/api-guide/concepts/pagination?context=linkedin/marketing/context
        """
        parsed_response = response.json()
        is_elements_less_than_limit = len(parsed_response.get("elements")) < self.records_limit

        # Note: The API might return fewer records than requested within the limits during pagination.
        # This behavior is documented at: https://github.com/airbytehq/airbyte/issues/34164
        paging_params = parsed_response.get("paging", {})
        is_end_of_records = (
            paging_params["total"] - paging_params["start"] <= self.records_limit
            if all(param in paging_params for param in ("total", "start"))
            else True
        )

        if is_elements_less_than_limit and is_end_of_records:
            return None
        return {"start": paging_params.get("start") + self.records_limit}


class Accounts(LinkedinAdsStream):
    """
    Get Accounts data. More info about LinkedIn Ads / Accounts:
    https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads/account-structure/create-and-manage-accounts?tabs=http&view=li-lms-2023-05#search-for-accounts
    """

    endpoint = "adAccounts"
    use_cache = True

    def request_headers(self, stream_state: Mapping[str, Any], **kwargs) -> Mapping[str, Any]:
        """
        If account_ids are specified as user's input from configuration,
        we must use MODIFIED header: {'X-RestLi-Protocol-Version': '2.0.0'}
        """
        headers = super().request_headers(stream_state, **kwargs)
        headers.update({"X-RestLi-Protocol-Version": "2.0.0"} if self.accounts else {})
        return headers

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        """
        Override request_params() to have the ability to accept the specific account_ids from user's configuration.
        If we have list of account_ids, we need to make sure that the request_params are encoded correctly,
        We will get HTTP Error 500, if we use standard requests.urlencode methods to parse parameters,
        so the urlencode(..., safe=":(),%") is used instead, to keep the values as they are.
        """
        params = super().request_params(stream_state, stream_slice, next_page_token)
        if self.accounts:
            # Construct the URN for each account ID
            accounts = [f"urn:li:sponsoredAccount:{account_id}" for account_id in self.config.get("account_ids")]

            # Join the URNs into a single string, separated by commas, and URL encode only this part
            encoded_accounts = quote(",".join(accounts), safe=",")

            # Insert the encoded account IDs into the overall structure, keeping colons and parentheses outside safe
            params["search"] = f"(id:(values:List({encoded_accounts})))"
        return urlencode(params, safe=":(),%")


class IncrementalLinkedinAdsStream(LinkedinAdsStream):
    cursor_field = "lastModified"

    @property
    def primary_slice_key(self) -> str:
        """
        Define the main slice_key from `slice_key_value_map`. Always the first element.
        EXAMPLE:
            in : {"k1": "v1", "k2": "v2", ...}
            out : "k1"
        """
        return list(self.parent_values_map.keys())[0]

    @property
    @abstractmethod
    def parent_stream(self) -> LinkedinAdsStream:
        """Defines the parent stream for slicing, the class object should be provided."""

    @property
    def state_checkpoint_interval(self) -> Optional[int]:
        """Define the checkpoint from the record output size."""
        return 100

    def get_updated_state(self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]) -> Mapping[str, Any]:
        current_stream_state = {self.cursor_field: self.config.get("start_date")} if not current_stream_state else current_stream_state
        return {self.cursor_field: max(latest_record.get(self.cursor_field), current_stream_state.get(self.cursor_field))}


class LinkedInAdsStreamSlicing(IncrementalLinkedinAdsStream, ABC):
    """
    This class stands for provide stream slicing for other dependent streams.
    :: `parent_stream` - the reference to the parent stream class,
        by default it's referenced to the Accounts stream class, as far as a majority of streams are using it.
    :: `parent_values_map` - key_value map for stream slices in a format: {<slice_key_name>: <key inside record>}
    :: `search_param` - the query param to pass with request_params
    """

    parent_stream = Accounts
    parent_values_map = {"account_id": "id"}

    def filter_records_newer_than_state(
        self, stream_state: Mapping[str, Any] = None, records_slice: Iterable[Mapping[str, Any]] = None
    ) -> Iterable:
        """For the streams that provide the cursor_field `lastModified`, we filter out the old records."""
        if stream_state:
            for record in records_slice:
                if record[self.cursor_field] >= stream_state.get(self.cursor_field):
                    yield record
        else:
            yield from records_slice

    def read_records(
        self, stream_state: Mapping[str, Any] = None, stream_slice: Optional[Mapping[str, Any]] = None, **kwargs
    ) -> Iterable[Mapping[str, Any]]:
        stream_state = stream_state or {}
        parent_stream = self.parent_stream(config=self.config)
        for record in parent_stream.read_records(**kwargs):
            child_stream_slice = super().read_records(stream_slice=get_parent_stream_values(record, self.parent_values_map), **kwargs)
            yield from self.filter_records_newer_than_state(stream_state=stream_state, records_slice=child_stream_slice)


class AccountUsers(OffsetPaginationMixin, LinkedInAdsStreamSlicing):
    """
    Get AccountUsers data using `account_id` slicing. More info about LinkedIn Ads / AccountUsers:
    https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads/account-structure/create-and-manage-account-users?tabs=http&view=li-lms-2023-05#find-ad-account-users-by-accounts
    """

    endpoint = "adAccountUsers"
    # Account_users stream doesn't have `id` property, so the "account" is used instead.
    primary_key = "account"
    search_param = "accounts"

    def request_params(self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, **kwargs) -> MutableMapping[str, Any]:
        params = super().request_params(stream_state=stream_state, **kwargs)
        params["q"] = self.search_param
        params["accounts"] = f"urn:li:sponsoredAccount:{stream_slice.get('account_id')}"  # accounts=
        return urlencode(params, safe="():,%")


class CampaignGroups(LinkedInAdsStreamSlicing):
    """
    Get CampaignGroups data using `account_id` slicing.
    More info about LinkedIn Ads / CampaignGroups:
    https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads/account-structure/create-and-manage-campaign-groups?tabs=http&view=li-lms-2023-05#search-for-campaign-groups
    """

    endpoint = "adCampaignGroups"

    def path(
        self,
        *,
        stream_state: Mapping[str, Any] = None,
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> str:
        return f"{self.parent_stream.endpoint}/{stream_slice.get('account_id')}/{self.endpoint}"

    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        headers = super().request_headers(stream_state, stream_slice, next_page_token)
        return headers | {"X-Restli-Protocol-Version": "2.0.0"}

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        params = super().request_params(stream_state, stream_slice, next_page_token)
        params["search"] = "(status:(values:List(ACTIVE,ARCHIVED,CANCELED,DRAFT,PAUSED,PENDING_DELETION,REMOVED)))"
        return urlencode(params, safe="():,%")


class Campaigns(LinkedInAdsStreamSlicing):
    """
    Get Campaigns data using `account_id` slicing.
    More info about LinkedIn Ads / Campaigns:
    https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads/account-structure/create-and-manage-campaigns?tabs=http&view=li-lms-2023-05#search-for-campaigns
    """

    endpoint = "adCampaigns"
    use_cache = True

    def path(
        self,
        *,
        stream_state: Mapping[str, Any] = None,
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> str:
        return f"{self.parent_stream.endpoint}/{stream_slice.get('account_id')}/{self.endpoint}"

    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        headers = super().request_headers(stream_state, stream_slice, next_page_token)
        return headers | {"X-Restli-Protocol-Version": "2.0.0"}

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        params = super().request_params(stream_state, stream_slice, next_page_token)
        params["search"] = "(status:(values:List(ACTIVE,PAUSED,ARCHIVED,COMPLETED,CANCELED,DRAFT,PENDING_DELETION,REMOVED)))"
        #
        return urlencode(params, safe="():,%")


class Creatives(LinkedInAdsStreamSlicing):
    """
    Get Creatives data using `campaign_id` slicing.
    More info about LinkedIn Ads / Creatives:
    https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads/account-structure/create-and-manage-creatives?tabs=http%2Chttp-update-a-creative&view=li-lms-2023-05#search-for-creatives
    """

    endpoint = "creatives"
    parent_stream = Accounts
    cursor_field = "lastModifiedAt"
    # standard records_limit=500 returns error 400: Request would return too many entities; https://github.com/airbytehq/oncall/issues/2159
    records_limit = 100

    def path(
        self,
        *,
        stream_state: Mapping[str, Any] = None,
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> str:
        return f"{self.parent_stream.endpoint}/{stream_slice.get('account_id')}/{self.endpoint}"

    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        headers = super().request_headers(stream_state, stream_slice, next_page_token)
        headers.update({"X-RestLi-Method": "FINDER"})
        return headers

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        params = super().request_params(stream_state, stream_slice, next_page_token)
        params.update({"q": "criteria"})
        return urlencode(params, safe="():,%")

    def get_updated_state(self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]) -> Mapping[str, Any]:
        current_stream_state = (
            {self.cursor_field: pendulum.parse(self.config.get("start_date")).format("x")}
            if not current_stream_state
            else current_stream_state
        )
        return {self.cursor_field: max(latest_record.get(self.cursor_field), int(current_stream_state.get(self.cursor_field)))}


class Conversions(OffsetPaginationMixin, LinkedInAdsStreamSlicing):
    """
    Get Conversions data using `account_id` slicing.
    https://learn.microsoft.com/en-us/linkedin/marketing/integrations/ads-reporting/conversion-tracking?view=li-lms-2023-05&tabs=curl#find-conversions-by-ad-account
    """

    endpoint = "conversions"
    search_param = "account"

    def request_headers(
        self, stream_state: Mapping[str, Any], stream_slice: Mapping[str, Any] = None, next_page_token: Mapping[str, Any] = None
    ) -> Mapping[str, Any]:
        headers = super().request_headers(stream_state, stream_slice, next_page_token)
        headers.update({"X-Restli-Protocol-Version": "2.0.0"})
        return headers

    def request_params(
        self,
        stream_state: Mapping[str, Any],
        stream_slice: Mapping[str, Any] = None,
        next_page_token: Mapping[str, Any] = None,
    ) -> MutableMapping[str, Any]:
        params = super().request_params(stream_state, stream_slice, next_page_token)
        params["q"] = self.search_param
        params["account"] = f"urn%3Ali%3AsponsoredAccount%3A{stream_slice.get('account_id')}"

        return urlencode(params, safe="():,%")

    def get_updated_state(self, current_stream_state: MutableMapping[str, Any], latest_record: Mapping[str, Any]) -> Mapping[str, Any]:
        current_stream_state = (
            {self.cursor_field: pendulum.parse(self.config.get("start_date")).format("x")}
            if not current_stream_state
            else current_stream_state
        )
        return {self.cursor_field: max(latest_record.get(self.cursor_field), int(current_stream_state.get(self.cursor_field)))}
